# SPDX-License-Identifier: GPL-2.0

import subprocess

def avail_cmd(cmd):
    try:
        subprocess.check_output(['which', cmd], stderr=subprocess.DEVNULL)
        return True
    except:
        return False



# SPDX-License-Identifier: GPL-2.0

import argparse

import _damon_args

def main(args):
    category = args.category
    parser = argparse.ArgumentParser(add_help=False)
    if category == 'monitoring':
        print('command line options for monitoring-part DAMON parameters')
        print()
        _damon_args.set_monitoring_argparser(parser, hide_help=False)
    elif category == 'damos':
        print('command line options for DAMOS parameters')
        print()
        _damon_args.set_damos_argparser(parser, hide_help=False)
    elif category == 'all':
        print('command line options for all DAMON parameters')
        print()
        _damon_args.set_misc_damon_params_argparser(parser)
    help_msg = parser.format_help()
    pars = help_msg.split('\n\n')
    # the first paragraph is usage for this virtual command.  Ignore.
    pars = pars[1:]
    if len(pars) == 1:
        print('\n'.join(pars[0].split('\n')[1:]))
    else:
        print('\n\n'.join(pars))
    if category == 'all':
        print(' '.join([
            "Also there are command line options for setting only monitoring",
            "and DAMOS-part DAMON parameters.",
            "Use 'damo help damon_param_options {monitoring,damos}'",
            "for those.",
            ]))

def set_argparser(parser):
    parser.add_argument(
            'category', choices=['all', 'monitoring', 'damos'],
            help='category of DAMON parameters to get help for')
    parser.description = 'help for DAMON parameters command line options'
    return parser



# SPDX-License-Identifier: GPL-2.0

import collections
import json

import _damo_yaml
import damo_report_access

def main(args):
    handled = damo_report_access.handle_ls_keywords(args)
    if handled:
        return
    fmt = damo_report_access.set_formats(args, records=[])
    kvpairs = fmt.to_kvpairs(raw=args.raw)

    if args.output_format == 'json':
        print(json.dumps(kvpairs, indent=4))
    elif args.output_format == 'yaml':
        text, err = _damo_yaml.dump(kvpairs)
        if err is not None:
            print('yaml dump failed (%s)' % err)
            exit(1)
        print(text)

def set_argparser(parser):
    parser.description = ' '.join([
        'format DAMON monitoring results visualization parameters'])
    damo_report_access.add_fmt_args(parser)
    parser.add_argument(
            '--output_format', choices=['json', 'yaml'], default='json',
            help='format of the output')
    parser.add_argument(
            '--raw', action='store_true',
            help='print numbers in machine friendly raw form')



# SPDX-License-Identifier: GPL-2.0

import os

import _damon

def main(args):
    _damon.ensure_root_permission()

    module_name = args.module_name
    param_dir = '/sys/module/damon_%s/parameters' % module_name
    if args.action == 'read':
        if args.parameter is not None:
            with open(os.path.join(param_dir, args.parameter), 'r') as f:
                print(f.read().strip())
        else:
            for param in os.listdir(param_dir):
                with open(os.path.join(param_dir, param), 'r') as f:
                    print('%s: %s' % (param, f.read().strip()))
    elif args.action == 'write':
        if len(args.parameter_value) % 2 != 0:
            print('wrong paramter_value')
            exit(1)

        for i in range(0, len(args.parameter_value), 2):
            param_name = args.parameter_value[i]
            param_val = args.parameter_value[i + 1]
            with open(os.path.join(param_dir, param_name), 'w') as f:
                f.write(param_val)

def set_argparser(parser):
    subparsers = parser.add_subparsers(
            title='action', dest='action', metavar='<action>')
    subparsers.required = True

    parser_read = subparsers.add_parser('read', help='read parameters')
    parser_read.add_argument(
            'parameter', metavar='<parameter name>', nargs='?',
            help='parameter to read.')

    parser_write = subparsers.add_parser('write', help='write parameters')
    parser_write.add_argument(
            'parameter_value', metavar=('<parameter name> <value>'), nargs='+',
            help='name of the parameter and the value to write')
    return parser



# SPDX-License-Identifier: GPL-2.0

"Print out the distribution of the memory footprint of the given trace"

import sys
import tempfile

import _damo_dist
import _damo_fmt_str
import _damo_records

def set_argparser(parser):
    parser.add_argument('metric', choices=['vsz', 'rss', 'sys_used', 'all'],
                        default='all', nargs='?',
                        help='memory footprint metric to show')
    parser.add_argument('--input', '-i', type=str, metavar='<file>',
                        default='damon.data.mem_footprint',
                        help='input file name')
    parser.add_argument('--range', '-r', type=int, nargs=3,
                        metavar=('<start>', '<stop>', '<step>'),
                        default=[0,101,25],
                        help='range of percentiles to print')
    parser.add_argument('--sortby', '-s', choices=['time', 'size'],
                        default='size',
                        help='the metric to sort the footprints for')
    parser.add_argument('--plot', '-p', type=str, metavar='<file>',
                        help='plot the distribution to an image file')
    parser.add_argument('--nr_cols_bar', type=int, metavar='<num>',
                        default=59,
                        help='max columns of output')
    parser.add_argument('--raw_number', action='store_true',
                        help='use machine-friendly raw numbers')
    parser.add_argument('--all_footprint', action='store_true',
                        help='print not percentiles but all footprint values')
    parser.description = 'Show distribution of memory footprint'

def get_dists(records, metric, do_sort):
    dists = []
    footprint_snapshots = _damo_records.load_mem_footprint(records)
    for snapshot in footprint_snapshots:
        footprint_bytes = 0
        for pid, fp in snapshot.footprints.items():
            if metric == 'sys_used':
                if pid is not None:
                    continue
                footprint_bytes = (fp.total - fp.free) * 1024
            # ignore SysMemFootprint
            if pid is None:
                continue
            # todo: get real page size of the system
            if metric == 'vsz':
                footprint_bytes += fp.size * 4096
            elif metric == 'rss':
                footprint_bytes += fp.resident * 4096
        dists.append(footprint_bytes)
    if do_sort:
        dists.sort()
    return dists

def main(args):
    if args.metric == 'all':
        for metric in ['vsz', 'rss', 'sys_used']:
            args.metric = metric
            main(args)
        return

    dists = get_dists(args.input, args.metric, args.sortby == 'size')

    percentiles = range(args.range[0], args.range[1], args.range[2])
    raw_number = args.raw_number
    if not args.plot:
        _damo_dist.pr_dists(
                args.metric, dists, percentiles, args.all_footprint,
                _damo_fmt_str.format_sz, raw_number, args.nr_cols_bar)
        return

    tmp_path = tempfile.mkstemp()[1]
    with open(tmp_path, 'w') as f:
        f.write(_damo_dist.fmt_dists(
            args.metric, dists, percentiles, args.all_footprint,
            _damo_fmt_str.format_sz, True, 0))

    sort_by_sz = True
    if args.sortby == 'time':
        sort_by_sz = False

    xlabel = 'runtime (percent)'
    if sort_by_sz:
        xlabel = 'percentile'
    err = _damo_dist.plot_dist(tmp_path, args.plot, xlabel,
            'memory footprint (kilobytes)')
    if err:
        print('plot failed (%s)' % err)



# SPDX-License-Identifier: GPL-2.0

import os

import _damo_subcmds
import damo_lru_sort
import damo_module_general
import damo_reclaim

subcmds = [
        _damo_subcmds.DamoSubCmd(
            name='reclaim', module=damo_reclaim, msg='DAMON_RECLAIM'),
        _damo_subcmds.DamoSubCmd(
            name='lru_sort', module=damo_lru_sort, msg='DAMON_LRU_SORT'),
        ]

def main(args):
    for subcmd in subcmds:
        if subcmd.name == args.module_name:
            subcmd.execute(args)

def set_subcmds(subcmds):
    for module in os.listdir('/sys/module'):
        if not module.startswith('damon_'):
            continue
        module_name = module[len('damon_'):]
        if module_name in ['reclaim', 'lru_sort']:
            continue
        subcmds.append(_damo_subcmds.DamoSubCmd(
            name=module_name, module=damo_module_general,
            msg='Control DAMON_%s' % module_name.upper()))

def set_argparser(parser):
    subparsers = parser.add_subparsers(
            title='DAMON module name', dest='module_name',
            metavar='<module name>',
            help='the name of DAMON module to control')
    subparsers.required = True

    set_subcmds(subcmds)

    for subcmd in subcmds:
        subcmd.add_parser(subparsers)



# SPDX-License-Identifier: GPL-2.0

"""
Print basic information of the access monitoring results record file.
"""

import _damo_fmt_str
import _damo_records

class GuideRegion:
    start_addr = None
    end_addr = None
    heats = None

    def heat_per_byte(self):
        if self.heats is None:
            return 0
        return float(self.heats) / (self.end_addr - self.start_addr)

    def __init__(self, start_addr, end_addr):
        self.start_addr = start_addr
        self.end_addr = end_addr

class GuideInfo:
    kdamond_idx = None
    context_idx = None
    scheme_idx = None
    tid = None
    start_time = None
    end_time = None
    lowest_addr = None
    highest_addr = None
    gaps = None
    contig_regions = None  # list of GuideRegion objects

    def __init__(self, kdamond_idx, context_idx, scheme_idx, tid, start_time):
        self.kdamond_idx = kdamond_idx
        self.context_idx = context_idx
        self.scheme_idx = scheme_idx
        self.tid = tid
        self.start_time = start_time
        self.gaps = []

    def regions(self):
        regions = []
        region = [self.lowest_addr]
        for gap in self.gaps:
            for idx, point in enumerate(gap):
                if idx == 0:
                    region.append(point)
                    regions.append(region)
                else:
                    region = [point]
        region.append(self.highest_addr)
        regions.append(region)
        return regions

    def total_space(self):
        ret = 0
        for r in self.regions():
            ret += r[1] - r[0]
        return ret

    def to_str(self, raw):
        lines = []
        if self.kdamond_idx is not None:
            lines.append('kdamond_idx: %s' % self.kdamond_idx)
        if self.context_idx is not None:
            line.sappend('context_idx: %s' % self.context_idx)
        if self.scheme_idx is not None:
            lines.append('scheme_idx: %s' % self.scheme_idx)
        if self.tid is not None:
            lines.append('target_id: %s' % self.tid)
        lines.append('time: [%s, %s) (%s)' % (
            _damo_fmt_str.format_time_ns(self.start_time, raw),
            _damo_fmt_str.format_time_ns(self.end_time, raw),
            _damo_fmt_str.format_time_ns(
                self.end_time - self.start_time, raw)))
        for idx, region in enumerate(self.contig_regions):
            lines.append('region\t%2d: [%s, %s) (%s)' % (
                idx,
                _damo_fmt_str.format_sz(region.start_addr, raw),
                _damo_fmt_str.format_sz(region.end_addr, raw),
                _damo_fmt_str.format_sz(
                    region.end_addr - region.start_addr, raw)))
        return '\n'.join(lines)

    def __str__(self):
        return self.to_str(raw=True)

def is_overlap(region1, region2):
    if region1[1] < region2[0]:
        return False
    if region2[1] < region1[0]:
        return False
    return True

def overlap_region_of(region1, region2):
    return [max(region1[0], region2[0]), min(region1[1], region2[1])]

def overlapping_regions(regions1, regions2):
    overlap_regions = []
    for r1 in regions1:
        for r2 in regions2:
            if is_overlap(r1, r2):
                r1 = overlap_region_of(r1, r2)
        if r1:
            overlap_regions.append(r1)
    return overlap_regions

def oldest_monitor_time(snapshot, record_intervals):
    if record_intervals is None:
        return snapshot.start_time
    longest_age = 0
    for region in snapshot.regions:
        if longest_age < region.age.aggr_intervals:
            longest_age = region.age.aggr_intervals
    aggr_interval_ns = record_intervals.aggr * 1000
    return snapshot.start_time - longest_age * aggr_interval_ns

def rec_to_guide_id(record):
    return '%s %s %s %s' % (record.kdamond_idx, record.context_idx,
                            record.scheme_idx, record.target_id)

def get_guide_info(records):
    "return the set of guide information for the moitoring result"
    guides = {}
    for record in records:
        guide_id = rec_to_guide_id(record)
        for snapshot in record.snapshots:
            if not guide_id in guides:
                guides[guide_id] = GuideInfo(
                        record.kdamond_idx, record.context_idx,
                        record.scheme_idx, record.target_id,
                        oldest_monitor_time(snapshot, record.intervals))
            guide = guides[guide_id]
            monitor_time = snapshot.end_time
            guide.end_time = monitor_time

            last_addr = None
            gaps = []
            for r in snapshot.regions:
                saddr = r.start
                eaddr = r.end

                if not guide.lowest_addr or saddr < guide.lowest_addr:
                    guide.lowest_addr = saddr
                if not guide.highest_addr or eaddr > guide.highest_addr:
                    guide.highest_addr = eaddr

                if not last_addr:
                    last_addr = eaddr
                    continue
                if last_addr != saddr:
                    gaps.append([last_addr, saddr])
                last_addr = eaddr

            if not guide.gaps:
                guide.gaps = gaps
            else:
                guide.gaps = overlapping_regions(guide.gaps, gaps)

    for guide_id, guide in guides.items():
        guide_regions = []
        for start, end in guide.regions():
            guide_regions.append(GuideRegion(start, end))
        guide.contig_regions = guide_regions

        for record in records:
            if rec_to_guide_id(record) != guide_id:
                continue
            for snapshot in record.snapshots:
                for region in snapshot.regions:
                    if region.nr_accesses.samples == 0:
                        continue
                    for gregion in guide.contig_regions:
                        if (region.end < gregion.start_addr or
                            gregion.end_addr < region.start):
                            continue
                        if gregion.heats is None:
                            gregion.heats = 0
                        gregion.heats += ((region.end - region.start) *
                                          region.nr_accesses.samples)

    return sorted(list(guides.values()), key=lambda x: x.total_space(),
                    reverse=True)

def pr_guide(records, raw_numbers=True):
    for guide in get_guide_info(records):
        print(guide.to_str(raw_numbers))

def main(args):
    records, err = _damo_records.get_records(record_file=args.input)
    if err != None:
        print('monitoring result file (%s) parsing failed (%s)' %
                (args.input, err))
        exit(1)
    pr_guide(records, args.raw_numbers)

def set_argparser(parser):
    parser.add_argument('--input', '-i', type=str, metavar='<file>', nargs='+',
            default='damon.data', help='input file name')
    parser.add_argument('--raw_numbers', action='store_true',
                        help='print numbers in raw format')



# SPDX-License-Identifier: GPL-2.0

"""
Transform DAMON monitoring results record into a heatmap.  The heatmap is
constructed with pixels that each shows when (time), which memory region
(space) was how frequently accessed (heat).  The time and space are represented
by the location of the pixel on the map, while the heat is represented by it's
color.

By default, the output shows the heatmap on the terminal.

If --output raw is given, the output shows the relative time, space, and heat
values of each pixel of the map on each line, like below.

    <time> <space> <heat>
    ...

By constructing the pixels based on the values, the user can draw more
human-readable heatmap.  gnuplot like plot tools can be used.  If '--heatmap'
option is given, this tool does that on behalf of the human when '--heatmap'
option is given.

If --output option is given with a file, the gnuplot-based heatmap image is
generated as the file.
"""

import os
import subprocess
import sys
import tempfile

import _damo_ascii_color
import _damo_fmt_str
import _damo_records
import damo_record_info
import damo_report_access

class HeatPixel:
    time = None
    addr = None
    heat = None

    def __init__(self, time, addr, heat):
        self.time = time
        self.addr = addr
        self.heat = heat

class HeatMap:
    time_start = None
    time_unit = None
    time_resol = None

    addr_start = None
    addr_unit = None
    addr_resol = None

    pixels = None # list of list of pixels

    def __init__(self, time_start, time_unit, time_resol, addr_start,
                 addr_unit, addr_resol):
        self.time_start = time_start
        self.time_unit = time_unit
        self.time_resol = time_resol

        self.addr_start = addr_start
        self.addr_unit = addr_unit
        self.addr_resol = addr_resol

        self.pixels = []
        for i in range(time_resol):
            self.pixels.append([])
            for j in range(addr_resol):
                pixel_time = time_start + i * time_unit
                pixel_addr = addr_start + j * addr_unit
                self.pixels[-1].append(
                        HeatPixel(int(pixel_time), int(pixel_addr), None))

    def pixels_idx_of_time(self, time_ns):
        return int((time_ns - self.time_start) / self.time_unit)

    def pixel_idx_of_addr(self, addr):
        return int((addr - self.addr_start) / self.addr_unit)

    def add_pixel_heat(self, pixels_idx, pixel_idx, region, snapshot, aggr_ns,
                       df_passed):
        pixel = self.pixels[pixels_idx][pixel_idx]
        observe_start_time = snapshot.start_time
        if aggr_ns is not None:
            observe_start_time -= region.age.aggr_intervals * aggr_ns

        account_time_start = max(observe_start_time, pixel.time)
        account_time_end = min(snapshot.end_time, pixel.time + self.time_unit)
        account_time = account_time_end - account_time_start

        account_addr_start = max(region.start, pixel.addr)
        account_addr_end = min(region.end, pixel.addr + self.addr_unit)
        account_sz = account_addr_end - account_addr_start

        nr_accesses = region.nr_accesses.samples
        if df_passed:
            nr_accesses = nr_accesses * region.sz_filter_passed / region.size()

        heat = nr_accesses * account_time * account_sz

        if pixel.heat is None:
            pixel.heat = 0

        pixel_time_space = self.time_unit * self.addr_unit

        heat += pixel.heat * pixel_time_space
        pixel.heat = float(heat) / pixel_time_space

    def pixels_idxs_range(self, region, snapshot, last_snapshot, aggr_ns):
        start_time = snapshot.start_time
        if aggr_ns is not None:
            start_time -= region.age.aggr_intervals * aggr_ns
        if last_snapshot is not None and start_time < last_snapshot.end_time:
            start_time = last_snapshot.end_time
        return range(
                self.pixels_idx_of_time(start_time),
                self.pixels_idx_of_time(snapshot.end_time) + 1)

    def add_heat(self, snapshot, last_snapshot, aggr_ns, df_passed):
        heatmap_addr_end = self.addr_start + self.addr_unit * self.addr_resol
        for region in snapshot.regions:
            if region.end < self.addr_start or heatmap_addr_end < region.start:
                continue

            for pixels_idx in self.pixels_idxs_range(
                    region, snapshot, last_snapshot, aggr_ns):
                if pixels_idx < 0 or self.time_resol <= pixels_idx:
                    continue
                for pixel_idx in range(
                        self.pixel_idx_of_addr(region.start),
                        self.pixel_idx_of_addr(region.end) + 1):
                    if pixel_idx < 0 or self.addr_resol <= pixel_idx:
                        continue
                    self.add_pixel_heat(
                            pixels_idx, pixel_idx, region, snapshot, aggr_ns,
                            df_passed)

    def highest_lowest_heats(self):
        highest = None
        lowest = None
        for row in self.pixels:
            for pixel in row:
                if pixel.heat is None:
                    continue
                if highest is None or highest < pixel.heat:
                    highest = pixel.heat
                if lowest is None or lowest > pixel.heat:
                    lowest = pixel.heat
        return highest, lowest

    def fmt_gnuplot_str(self, abs_time, abs_addr):
        lines = []
        for row in self.pixels:
            for pixel in row:
                time = pixel.time
                addr = pixel.addr
                if not abs_time:
                    time -= self.time_start
                if not abs_addr:
                    addr -= self.addr_start

                heat = pixel.heat if pixel.heat is not None else 'NaN'
                lines.append('%s\t%s\t%s' % (time, addr, heat))
        return '\n'.join(lines)

    def fmt_ascii_lines_map(self, colorset):
        lines = []
        # add pixels
        highest_heat, lowest_heat = self.highest_lowest_heats()
        if highest_heat is None and lowest_heat is None:
            return lines
        heat_unit = float(highest_heat + 1 - lowest_heat) / 9
        pixels = self.pixels
        for snapshot in pixels:
            chars = []
            for pixel in snapshot:
                if pixel.heat is None:
                    chars.append('%s ' %
                                 _damo_ascii_color.color_mode_start_txt(
                                     colorset, 4))
                    continue
                heat = int(float(pixel.heat - lowest_heat) / heat_unit)
                heat = min(heat, _damo_ascii_color.max_color_level())
                chars.append('%s%d' %
                        (_damo_ascii_color.color_mode_start_txt(colorset, heat),
                            heat))
            lines.append(''.join(chars) + _damo_ascii_color.color_mode_end_txt())
        return lines

    def fmt_ascii_lines_comments(self, colorset, print_colorset):
        lines = []
        if print_colorset:
            lines.append('# access_frequency: %s' %
                    _damo_ascii_color.color_samples(colorset))

        addr_start = self.addr_start
        addr_len = self.addr_unit * self.addr_resol
        addr_end = addr_start + addr_len
        lines.append('# x-axis: space [%s, %s) (%s)' % (
            _damo_fmt_str.format_sz(addr_start, False),
            _damo_fmt_str.format_sz(addr_end, False),
            _damo_fmt_str.format_sz(addr_len, False)))

        time_start = self.time_start
        time_len = self.time_unit * self.time_resol
        time_end = time_start + time_len
        lines.append('# y-axis: time [%s, %s) (%s)' % (
            _damo_fmt_str.format_time_ns(time_start, False),
            _damo_fmt_str.format_time_ns(time_end, False),
            _damo_fmt_str.format_time_ns(time_len, False)))
        pixels = self.pixels
        lines.append('# resolution: %dx%d (%s and %s for each character)' % (
            len(pixels[1]), len(pixels),
            _damo_fmt_str.format_sz(float(addr_len) / len(pixels[1]), False),
            _damo_fmt_str.format_time_ns(
                float(time_len) / len(pixels), False)))
        return lines

    def fmt_ascii_str(self, colorset, print_colorset):
        lines = self.fmt_ascii_lines_map(colorset)
        lines += self.fmt_ascii_lines_comments(colorset, print_colorset)
        return '\n'.join(lines)

def heatmap_from_records(
        records, time_range, addr_range, resols, df_passed):
    time_start, time_end = time_range
    addr_start, addr_end = addr_range
    time_resol, addr_resol = resols
    time_unit = (time_end - time_start) // time_resol
    addr_unit = (addr_end - addr_start) // addr_resol

    heatmap = HeatMap(time_start, time_unit, time_resol,
                      addr_start, addr_unit, addr_resol)
    last_snapshot = None
    for record in records:
        aggr_ns = None
        if record.intervals is not None:
            aggr_ns = record.intervals.aggr * 1000
        for snapshot in record.snapshots:
            heatmap.add_heat(snapshot, last_snapshot, aggr_ns, df_passed)
            last_snapshot = snapshot
    return heatmap

def mk_heatmap(args, address_range_idx, __records):
    records = []
    for record in __records:
        if record.kdamond_idx != args.kdamond_idx:
            continue
        if record.context_idx != args.context_idx:
            continue
        if record.scheme_idx != args.scheme_idx:
            continue
        if record.target_id != args.tid:
            continue
        records.append(record)

    return heatmap_from_records(
            records, args.time_range, args.address_range[address_range_idx],
            args.resol, args.df_passed)

def fmt_heats(args, address_range_idx, __records):
    heatmap = mk_heatmap(args, address_range_idx, __records)
    if args.output == 'stdout':
        return heatmap.fmt_ascii_str(
                args.stdout_colorset, not args.stdout_skip_colorset_example)

    return heatmap.fmt_gnuplot_str(args.abs_time, args.abs_addr)

def complete_time_range(user_input, guide):
    if user_input is None:
        return guide, None
    if not len(user_input) in [2, 3]:
        return None, 'wrong number of arguments'
    time_range = [_damo_fmt_str.text_to_ns(x) for x in user_input[:2]]
    if len(user_input) == 2:
        return time_range, None
    base_input = user_input[2]
    if base_input == 'guided':
        base_time = guide[0]
    else:
        base_time = _damo_fmt_str.text_to_ns(user_input[2])
    return [base_time + x for x in time_range], None

def get_guided_address_ranges(draw_range, guide):
    if draw_range == 'hottest':
        hottest_contig_region = sorted(
                guide.contig_regions, key=lambda x: x.heat_per_byte(),
                reverse=True)[0]
        return [[hottest_contig_region.start_addr,
                 hottest_contig_region.end_addr]]
    elif draw_range == 'all':
        return [[r.start_addr, r.end_addr]
                for r in guide.contig_regions]

def complete_address_range_one(user_input, guided):
    if not len(user_input) in [2, 3]:
        return None, 'wrong number of fiellds'
    address_range = [_damo_fmt_str.text_to_bytes(x) for x in user_input[:2]]
    if len(user_input) == 2:
        return address_range, None
    base_input = user_input[2]
    if base_input == 'guided':
        if guided is None:
            return None, 'guide not exists'
        base_addr = guided[0]
    else:
        base_addr = _damo_fmt_str.text_to_bytes(base_input)
    return [base_addr + x for x in address_range], None

def complete_address_range(user_input, draw_range, guide):
    guided_address_ranges = get_guided_address_ranges(draw_range, guide)
    if user_input is None:
        return guided_address_ranges, None
    else:
        address_ranges = []
        for idx, address_range in enumerate(user_input):
            if idx < len(guided_address_ranges):
                guided_range = guided_address_ranges[idx]
            else:
                guided_range = None
            ar, err = complete_address_range_one(address_range, guided_range)
            if err is not None:
                return None, err
            address_ranges.append(ar)
        return address_ranges, None

def complete_src_args(args, records):
    if (args.kdamond_idx is not None and args.context_idx is not None and
        args.scheme_idx is not None and args.tid is not None and
        args.time_range is not None and args.address_range is not None):
        return None
    guides = damo_record_info.get_guide_info(records)
    guide = guides[0]
    if args.kdamond_idx is None:
        args.kdamond_idx = guide.kdamond_idx
    if args.context_idx is None:
        args.context_idx = guide.context_idx
    if args.scheme_idx is None:
        args.scheme_idx = guide.scheme_idx
    if args.tid is None:
        args.tid = guide.tid

    proper_guides = [g for g in guides if
                g.kdamond_idx == args.kdamond_idx and
                g.context_idx == args.context_idx and g.scheme_idx ==
                args.scheme_idx and g.tid == args.tid]
    if len(proper_guides) == 0:
        return 'no proper guide'
    guide = proper_guides[0]

    args.time_range, err = complete_time_range(
            args.time_range, [guide.start_time, guide.end_time])
    if err is not None:
            return 'time range completion fail (%s)' % err

    args.address_range, err = complete_address_range(
            args.address_range, args.draw_range, guide)

    if not args.address_range:
        if args.draw_range == 'hottest':
            hottest_contig_region = sorted(
                    guide.contig_regions, key=lambda x: x.heat_per_byte(),
                    reverse=True)[0]
            args.address_range = [[hottest_contig_region.start_addr,
                                  hottest_contig_region.end_addr]]
        elif args.draw_range == 'all':
            args.address_range = [
                    [r.start_addr, r.end_addr] for r in guide.contig_regions]
    else:
        for idx, address_range in enumerate(args.address_range):
            args.address_range[idx] = [
                    _damo_fmt_str.text_to_bytes(x) for x in address_range]

    return None

def sort_regions_by_temperature_for_range(snapshot, record, start, end,
                                          weights):
    sorted_regions = []
    for region in snapshot.regions:
        if end < region.start or region.end < start:
            continue
        if region.start < start:
            region.start = start
        if end < region.end:
            region.end = end
        region.nr_accesses.add_unset_unit(record.intervals)
        region.age.add_unset_unit(record.intervals)
        sorted_regions.append(region)
    sorted_regions.sort(key=lambda r: damo_report_access.temperature_of(
        r, weights))
    for idx, region in enumerate(sorted_regions):
        if idx == 0:
            adjusted_start = start
        else:
            adjusted_start = sorted_regions[idx - 1].end
        adjusted_end = adjusted_start + region.size()
        region.start = adjusted_start
        region.end = adjusted_end
    return sorted_regions

def sort_regions_by_temperature(records, args):
    for record in records:
        for snapshot in record.snapshots:
            sorted_regions = []
            for start, end in args.address_range:
                sorted_regions += sort_regions_by_temperature_for_range(
                        snapshot, record, start, end,
                        [0] + args.temperature_weights)
            snapshot.regions = sorted_regions

def plot_range(orig_range, use_absolute_val):
    plot_range = [x for x in orig_range]
    if not use_absolute_val:
        plot_range[0] -= orig_range[0]
        plot_range[1] -= orig_range[0]
    return plot_range

def plot_heatmap(data_file, output_file, args, address_range, range_idx,
                 highest_heat, lowest_heat):
    terminal = output_file.split('.')[-1]
    if not terminal in ['pdf', 'jpeg', 'png', 'svg']:
        os.remove(data_file)
        print("Unsupported plot output type.")
        exit(-1)
    if range_idx > 0:
        tokens = output_file.split('.')
        tokens.insert(-1, '%d' % range_idx)
        output_file = '.'.join(tokens)

    x_range = plot_range(args.time_range, args.abs_time)
    y_range = plot_range(address_range, args.abs_addr)

    ylabel = 'Address (bytes)'
    if args.sort_temperature:
        ylabel = 'Size (bytes)'

    gnuplot_cmd = """
    set term %s;
    set output '%s';
    set key off;
    set xrange [%f:%f];
    set yrange [%f:%f];
    set cbrange [%f:%f];
    set xlabel 'Time (ns)';
    set ylabel '%s';
    plot '%s' using 1:2:3 with image;""" % (
            terminal, output_file, x_range[0], x_range[1],
            y_range[0], y_range[1], highest_heat, lowest_heat, ylabel,
            data_file)
    try:
        subprocess.call(['gnuplot', '-e', gnuplot_cmd])
    except Exception as e:
        print('executing gnuplot failed (%s)' % e)
    os.remove(data_file)

def mk_show_heatmap(records, args):
    heats_list = []
    for idx in range(len(args.address_range)):
        heatmap = mk_heatmap(args, idx, records)

        if args.output == 'stdout':
            print(heatmap.fmt_ascii_str(
                args.stdout_colorset, not args.stdout_skip_colorset_example))
            continue
        else:
            gnuplot_data_str = heatmap.fmt_gnuplot_str(
                    args.abs_time, args.abs_addr)
            if args.output == 'raw':
                print(gnuplot_data_str)
                continue
            # use gnuplot-based image plot
            highest_heat, lowest_heat = heatmap.highest_lowest_heats()
            tmp_path = tempfile.mkstemp()[1]
            with open(tmp_path, 'w') as f:
                f.write(gnuplot_data_str)
            plot_heatmap(
                    tmp_path, args.output, args, args.address_range[idx], idx,
                    highest_heat, lowest_heat)

def scale_range(start_end, ratio):
    start, end = start_end
    size = end - start
    new_size = size * ratio
    mid = int(start + size / 2)
    return [mid - new_size / 2, mid + new_size / 2]

def add_diff_range(start_end, ratio):
    start, end = start_end
    size = end - start
    diff = size * ratio
    return [start + diff, end + diff]

edit_zoom = 1
edit_scroll = 2
edit_time = 1
edit_space = 2
edit_time_space = 3

def edit_time_address_ranges(action, target, ratio, args):
    if action == edit_zoom and target in [edit_time, edit_time_space]:
        args.time_range = scale_range(args.time_range, ratio)
    if action == edit_zoom and target in [edit_space, edit_time_space]:
        for idx, start_end in enumerate(args.address_range):
            args.address_range[idx] = scale_range(start_end, ratio)
    if action == edit_scroll and target in [edit_time, edit_time_space]:
        args.time_range = add_diff_range(args.time_range, ratio)
    if action == edit_scroll and target in [edit_space, edit_time_space]:
        for idx, start_end in enumerate(args.address_range):
            args.address_range[idx] = add_diff_range(
                    args.address_range[idx], ratio)

def handle_shortcuts(answer, args):
    if not answer in ['+', '-', 'j', 'h', 'k', 'l']:
        return False
    if args.output == 'stdout':
        xaxis = edit_space
        yaxis = edit_time
        # on stdout, y-axis increases towards bottom
        down = -0.1
        up = 0.1
    else:
        xaxis = edit_time
        yaxis = edit_space
        # on none-stdout, y-axis increases towards top
        down = 0.1
        up = -0.1
    if answer == '+':
        edit_time_address_ranges(edit_zoom, edit_time_space, 0.9, args)
    elif answer == '-':
        edit_time_address_ranges(edit_zoom, edit_time_space, 1.1, args)
    elif answer == 'h':
        edit_time_address_ranges(edit_scroll, xaxis, 0.1, args)
    elif answer == 'j':
        edit_time_address_ranges(edit_scroll, yaxis, down, args)
    elif answer == 'k':
        edit_time_address_ranges(edit_scroll, yaxis, up, args)
    elif answer == 'l':
        edit_time_address_ranges(edit_scroll, xaxis, -0.1, args)
    return True

def print_interactive_edit_help():
    help_msg = [
            'Zoom or scroll using the prompt menu.',
            '',
            'Below inputs are supported.',
            '0,q,quit: Quit this mode.',
            '1,zoom: Continue prompts for zooming in/out the map.',
            '2,scroll: Continue prompts for scorrling the map.',
            '3,help: Print this help message.',
            'h: Scroll the map left, 10%.',
            'j: Scroll the map down, 10%.',
            'k: Scroll the map up, 10%.',
            'l: Scroll the map right, 10%.',
            '+: Zoom the map in, 10%.',
            '-: Zomm the map out, 10%.',
            '',
            ]
    print('\n'.join(help_msg))

def do_interactive_edit(args):
    if not args.interactive_edit:
        return True

    while True:
        answer = input('Enter (zoom, scroll, quit, help): ')
        handled = handle_shortcuts(answer, args)
        if handled:
            return False

        if answer in ['0', 'q', 'quit']:
            return True
        elif answer in ['3', 'help']:
            print_interactive_edit_help()
            continue
        elif answer in ['1', 'zoom']:
            action = edit_zoom
        elif answer in ['2', 'scroll']:
            action = edit_scroll
        else:
            print('Wrong input.')
            continue
        break
    answer = input('Enter (1. Time, 2. Space, 3. Time and Space): ')
    if anaswer == '1':
        target = edit_time
    if answer == '2':
        target = edit_space
    if answer == '3':
        target = edit_time_space
    ratio = float(input('Enter percentage: ')) / 100
    edit_time_address_ranges(action, target, ratio, args)
    return False

def main(args):
    records, err = _damo_records.get_records(record_file=args.input)
    if err != None:
        print('monitoring result file (%s) parsing failed (%s)' %
                (args.input, err))
        exit(1)

    # Use 80x40 or 500x500 resolution as default for stdout or image plots
    if args.resol is None:
        if args.output == 'stdout':
            args.resol = [40, 80]
        else:
            args.resol = [500, 500]

    if args.guide:
        damo_record_info.pr_guide(records)
        return
    if args.guide_human is True:
        damo_record_info.pr_guide(records, raw_numbers=False)
        return

    err = complete_src_args(args, records)
    if err is not None:
        print('source arguments completion fail (%s)' % err)
        exit(1)
    if args.sort_temperature:
        sort_regions_by_temperature(records, args)

    interactive_zoom_done = False
    while not interactive_zoom_done:
        mk_show_heatmap(records, args)
        interactive_zoom_done = do_interactive_edit(args)

def set_argparser(parser):
    parser.add_argument('--output', metavar='<output>', default='stdout',
                        help=' '.join(
                            ['output heatmap to generate.',
                             'can be a pdf/png/jpeg/svg file or',
                             'special keywords (\'stdout\', \'raw\')']))
    parser.add_argument('--input', '-i', type=str, metavar='<file>', nargs='+',
            default=['damon.data'], help='input file name')

    parser.add_argument('--kdamond_idx', metavar='<int>', type=int,
                        help='kdamond idx of record to print heatmap for')
    parser.add_argument('--context_idx', metavar='<int>', type=int,
                        help='context idx of record to print heatmap for')
    parser.add_argument('--scheme_idx', metavar='<int>', type=int,
                        help='scheme idx of record to print heatmap for')
    parser.add_argument('--tid', metavar='<id>', type=int,
                        help='target id of record to print heatmap for')
    parser.add_argument('--resol', metavar='<resolution>', type=int, nargs=2,
            help='resolutions for time and address axes')
    parser.add_argument('--time_range', metavar='<nanoseconds>', nargs='+',
            help='"<start> <end> [base]" of time ranges for the heamap.' \
                    '[base] can be "guided" for the guided start time.'
                    )
    parser.add_argument('--draw_range', choices=['hottest', 'all'],
                        default='hottest',
                        help='which ranges to draw heatmap for')
    parser.add_argument(
            '--address_range', metavar='<address>', nargs='+', action='append',
            help='"<start> <end> [base]" of address ranges for the heatmap.' \
                    '[base] can be "guided" for the guided start address.')
    parser.add_argument('--abs_time', action='store_true', default=False,
            help='display absolute time in output')
    parser.add_argument('--abs_addr', action='store_true', default=False,
            help='display absolute address in output')

    parser.add_argument('--guide', action='store_true',
            help='print a guidance for the ranges and resolution settings')
    parser.add_argument('--guide_human', action='store_true',
                        help='print the guidance in human-friendly format')
    parser.add_argument('--stdout_colorset', default='gray',
            choices=['gray', 'flame', 'emotion'],
            help='color theme for access frequencies')
    parser.add_argument('--stdout_skip_colorset_example',
            action='store_true',
            help='skip printing example colors at the output')
    parser.add_argument('--df_passed', action='store_true',
                        help='show heatmap for only DAMOS filter-passed parts')
    parser.add_argument('--sort_temperature', action='store_true',
                        help='sort regions by temperature on space')
    parser.add_argument('--temperature_weights', nargs=2, type=float,
                        metavar=('<frequency>', '<age>'), default=[1.0, 1.0],
                        help='temperature calculation weights')
    parser.add_argument('--interactive_edit', action='store_true',
                        help='interactively zoom and scroll the maps')
    parser.description = 'Show when which address ranges were how frequently accessed'



# SPDX-License-Identifier: GPL-2.0

"""
Provides manual for given topics.
"""

import _damo_subcmds
import damo_help_access_filter_options
import damo_help_access_format_options
import damo_help_damon_param_options

subcmds = [
        _damo_subcmds.DamoSubCmd(
            name='damon_param_options', module=damo_help_damon_param_options,
            msg='DAMON parameter command line options'),
        _damo_subcmds.DamoSubCmd(
            name='access_filter_options',
            module=damo_help_access_filter_options,
            msg="DAMON monitoring results filtering options"),
        _damo_subcmds.DamoSubCmd(
            name='access_format_options',
            module=damo_help_access_format_options,
            msg="DAMON monitoring results visualization options"),
        ]

def main(args):
    for subcmd in subcmds:
        if subcmd.name == args.topic:
            subcmd.execute(args)

def set_argparser(parser):
    subparsers = parser.add_subparsers(
            title='topic', dest='topic', metavar='<topic>',
            help='topic to get help for')
    subparsers.required = True

    for subcmd in subcmds:
        subcmd.add_parser(subparsers)



# SPDX-License-Identifier: GPL-2.0

import platform

def format_nr_positive(nr, machine_friendly):
    raw_string = '%d' % nr
    if machine_friendly:
        return raw_string
    fields = []
    for i in range(0, len(raw_string), 3):
        start_idx = max(0, len(raw_string) - i - 3)
        end_idx = len(raw_string) - i
        fields = [raw_string[start_idx:end_idx]] + fields
    return ','.join(fields)

def format_nr(nr, machine_friendly):
    if nr < 0:
        return '-' + format_nr_positive(nr * -1, machine_friendly)
    return format_nr_positive(nr, machine_friendly)

def format_sz(sz_bytes, machine_friendly):
    if machine_friendly:
        return '%d' % sz_bytes

    sz_bytes = float(sz_bytes)
    if sz_bytes == float(ulong_max):
        return 'max'
    if sz_bytes > 1<<60:
        return '%.3f EiB' % (sz_bytes / (1<<60))
    if sz_bytes > 1<<50:
        return '%.3f PiB' % (sz_bytes / (1<<50))
    if sz_bytes > 1<<40:
        return '%.3f TiB' % (sz_bytes / (1<<40))
    if sz_bytes > 1<<30:
        return '%.3f GiB' % (sz_bytes / (1<<30))
    if sz_bytes > 1<<20:
        return '%.3f MiB' % (sz_bytes / (1<<20))
    if sz_bytes > 1<<10:
        return '%.3f KiB' % (sz_bytes / (1<<10))
    return '%d B' % sz_bytes

def format_addr_range(start, end, machine_friendly):
    return '[%s, %s) (%s)' % (
            format_nr(start, machine_friendly),
            format_nr(end, machine_friendly),
            format_sz(end - start, machine_friendly))

ns_ns = 1
us_ns = 1000
ms_ns = 1000 * us_ns
sec_ns = 1000 * ms_ns
minute_ns = 60 * sec_ns
hour_ns = 60 * minute_ns
day_ns = 24 * hour_ns

nsecs_to_unit = {1: 'ns', us_ns: 'us', ms_ns: 'ms', sec_ns: 's',
        minute_ns: 'm', hour_ns: 'h', day_ns: 'd'}

def format_time_ns_min_unit(time_ns, min_unit, machine_friendly):
    if machine_friendly:
        return '%d' % time_ns

    if time_ns < 0:
        sign = '-'
        time_ns = -1 * time_ns
    else:
        sign = ''

    if time_ns >= ulong_max:
        return '%smax' % sign

    for unit_nsecs in sorted(nsecs_to_unit.keys(), reverse=True):
        if time_ns < unit_nsecs:
            continue
        if unit_nsecs == min_unit:
            if time_ns % unit_nsecs:
                return '%s%.3f %s' % (sign, time_ns / unit_nsecs,
                        nsecs_to_unit[unit_nsecs])
            else:
                return '%s%d %s' % (sign, time_ns / unit_nsecs,
                        nsecs_to_unit[unit_nsecs])

        unit_nr = int(time_ns / unit_nsecs)
        unit_str = '%s %s' % (
                format_nr(unit_nr, False), nsecs_to_unit[unit_nsecs])

        less_unit_ns = time_ns - unit_nr * unit_nsecs
        if less_unit_ns == 0:
            return '%s%s' % (sign, unit_str)
        else:
            return '%s%s %s' % (
                    sign, unit_str, format_time_ns_min_unit(
                        less_unit_ns, min_unit, False))
    return '0 ns'

def format_time_ns_exact(time_ns, machine_friendly):
    return format_time_ns_min_unit(time_ns, ns_ns, machine_friendly)

def format_time_us_exact(time_us, machine_friendly):
    return format_time_ns_exact(time_us * us_ns, machine_friendly)

def format_time_ms_exact(time_ms, machine_friendly):
    return format_time_ns_exact(time_ms * ms_ns, machine_friendly)

def format_time_ns(time_ns, machine_friendly):
    if machine_friendly:
        return '%d' % time_ns

    time_ns = float(time_ns)
    if time_ns < 0:
        sign = '-'
        time_ns = time_ns * -1
    else:
        sign = ''
    if time_ns >= ulong_max:
        return '%smax' % sign

    if time_ns >= hour_ns:
        hour = int(time_ns / hour_ns)
        hour_str = '%d h' % hour

        less_hour_ns = time_ns - (hour * hour_ns)
        if less_hour_ns == 0:
            return '%s%s' % (sign, hour_str)
        return '%s%s %s' % (
                sign, hour_str, format_time_ns(less_hour_ns, False))
    if time_ns >= minute_ns:
        if time_ns % minute_ns == 0:
            return '%s%d m' % (sign, time_ns / minute_ns)
        if time_ns % sec_ns == 0:
            return '%s%d m %d s' % (sign, time_ns / minute_ns,
                    (time_ns % minute_ns) / sec_ns)
        return '%s%d m %.3f s' % (sign, time_ns / minute_ns,
                (time_ns % minute_ns) / sec_ns)
    if time_ns >= sec_ns:
        if time_ns % sec_ns == 0:
            return '%s%d s' % (sign, time_ns / sec_ns)
        return '%s%.3f s' % (sign, time_ns / sec_ns)
    if time_ns >= ms_ns:
        if time_ns % ms_ns == 0:
            return '%s%d ms' % (sign, time_ns / ms_ns)
        return '%s%.3f ms' % (sign, time_ns / ms_ns)
    if time_ns >= us_ns:
        if time_ns % us_ns == 0:
            return '%s%d us' % (sign, time_ns / us_ns)
        return '%s%.3f us' % (sign, time_ns / us_ns)
    return '%s%d ns' % (sign, time_ns)

def format_time_us(time_us, machine_friendly):
    if machine_friendly:
        return '%d' % time_us

    return format_time_ns(time_us * 1000, machine_friendly)

def format_time_ms(time_ms, machine_friendly):
    if machine_friendly:
        return '%d' % time_ms

    return format_time_ns(time_ms * 1000000, machine_friendly)

def format_time_sec(time_sec, machine_friendly):
    if machine_friendly:
        return '%d' % time_sec

    return format_time_ns(time_sec * 1000000000, machine_friendly)

def format_ratio(ratio, machine_friendly):
    if machine_friendly:
        return '%f' % ratio

    over_percent = int(ratio * 100)
    over_percent_str = format_nr(over_percent, machine_friendly)

    under_percent = ratio * 100 % 1
    under_percent_str = ('%.7f' % under_percent).rstrip('0')
    # cut '0.' prefix
    under_percent_str = under_percent_str[2:]

    if under_percent_str == '':
        return '%s %%' % over_percent_str
    return '%s.%s %%' % (over_percent_str, under_percent_str)

def format_percent(percent, machine_friendly):
    if machine_friendly:
        return '%f' % percent

    return format_ratio(float(percent) / 100, machine_friendly)

def format_permil(permil, machine_friendly):
    if machine_friendly:
        return '%f' % permil

    return format_ratio(float(permil) / 1000, machine_friendly)

def format_bp(bp, machine_friendly):
    if machine_friendly:
        return '%f' % bp

    return format_ratio(float(bp) / 10000, machine_friendly)

def indent_lines(string, indent_width):
    return '\n'.join([' ' * indent_width + l for l in string.split('\n')])

number_types = [int, float]

try:
    # for python2
    number_types.append(long)
except:
    pass

uint_max = 2**32 - 1
ulong_max = 2**64 - 1
if platform.architecture()[0] != '64bit':
    ulong_max = 2**32 - 1

unit_to_bytes = {'B': 1,
        'K': 1 << 10, 'KB': 1 << 10, 'KiB': 1 << 10,
        'M': 1 << 20, 'MB': 1 << 20, 'MiB': 1 << 20,
        'G': 1 << 30, 'GB': 1 << 30, 'GiB': 1 << 30,
        'T': 1 << 40, 'TB': 1 << 40, 'TiB': 1 << 40,
        'P': 1 << 50, 'PB': 1 << 50, 'PiB': 1 << 50,
        'E': 1 << 60, 'EB': 1 << 60, 'EiB': 1 << 60}

def text_to_nr(txt):
    if type(txt) in number_types:
        return txt

    if txt == 'max':
        return ulong_max

    new_txt = ''.join([c for c in txt if c != ','])
    try:
        return int(new_txt)
    except:
        pass
    return float(new_txt)

def try_common_input(txt, min_val=0, max_val=ulong_max):
    'return success and number'
    if txt == 'min':
        return True, min_val
    if txt == 'max':
        return True, max_val
    try:
        return True, text_to_nr(txt)
    except:
        pass
    return False, None

def text_to_bytes(txt):
    success, number = try_common_input(txt)
    if success:
        return number

    unit = None
    if len(txt) > 3:
        unit = txt[len(txt) - 3:]
        if unit in unit_to_bytes:
            number = text_to_nr(txt[:-3])
        else:
            unit = None

    if unit == None:
        if len(txt) > 2:
            unit = txt[len(txt) - 2:]
            if unit in unit_to_bytes:
                number = text_to_nr(txt[:-2])
            else:
                unit = None

    if unit == None:
        if txt[-1] in unit_to_bytes:
            unit = txt[-1]
            number = text_to_nr(txt[:-1])
        else:
            unit = 'B'
            number  = text_to_nr(txt)

    return min(ulong_max, int(number * unit_to_bytes[unit]))

unit_to_nsecs = {'ns': ns_ns, 'us': us_ns, 'ms': ms_ns, 's': sec_ns,
        'm': minute_ns, 'h': hour_ns, 'd': day_ns}

def text_to_ns(txt):
    success, number = try_common_input(txt)
    if success:
        return number

    fields = txt.split()
    if len(fields) > 1:
        result_us = 0
        for i in range(0, len(fields), 2):
            result_us += text_to_ns(''.join(fields[i: i + 2]))
        return result_us

    if not txt[-2:] in unit_to_nsecs and not txt[-1] in unit_to_nsecs:
        return float(txt)

    unit = txt[-2:]
    if unit in ['ns', 'us', 'ms']:
        number = text_to_nr(txt[:-2])
    else:
        unit = txt[-1]
        number = text_to_nr(txt[:-1])
    return number * unit_to_nsecs[unit]

def text_to_us(txt):
    success, number = try_common_input(txt)
    if success:
        return number

    return text_to_ns(txt) / us_ns

def text_to_ms(txt):
    success, number = try_common_input(txt)
    if success:
        return number

    return text_to_us(txt) / 1000

def text_to_sec(txt):
    success, number = try_common_input(txt)
    if success:
        return number

    return text_to_ms(txt) / 1000

def text_to_ratio(txt):
    success, number = try_common_input(txt, 0.0)
    if success:
        return number

    is_percent = False
    if txt[-1] == '%':
        is_percent = True
        txt = txt[:-1]
    ratio = text_to_nr(txt)
    if is_percent:
        ratio /= 100.0
    return ratio

def text_to_bp(txt):
    success, number = try_common_input(txt, 0)
    if success:
        return number

    return text_to_ratio(txt) * 10000

def text_to_permil(txt):
    success, number = try_common_input(txt, 0)
    if success:
        return number

    return text_to_ratio(txt) * 1000

def text_to_percent(txt):
    success, number = try_common_input(txt, 0)
    if success:
        return number

    return text_to_ratio(txt) * 100

def text_to_nr_unit(txt):
    fields = txt.split()
    if len(fields) != 2:
        raise Exception('text_to_nr_unit requires two fields')
    return text_to_nr(fields[0]), fields[1]

def text_to_bool(txt):
    if type(txt) == bool:
        return txt

    true_txts = ['y', 'yes', 'true']
    false_txts = ['n', 'no', 'false']

    txt = txt.lower()
    if txt in true_txts:
        return True
    elif txt in false_txts:
        return False
    else:
        raise Exception('txt should be one of %s but %s' %
                (' '.join(true_txts + false_txts), txt))



# SPDX-License-Identifier: GPL-2.0

import argparse

import _damo_records

def main(args):
    print('command line options for monitoring results visualization options')
    parser = argparse.ArgumentParser(add_help=False)
    _damo_records.set_filter_argparser(parser, hide_help=False)
    help_msg = parser.format_help()
    pars = help_msg.split('\n\n')
    # the first paragraph is usage for this virtual command.  Ignore.
    pars = pars[1:]
    if len(pars) == 1:
        print('\n'.join(pars[0].split('\n')[1:]))
    else:
        print('\n\n'.join(pars))

def set_argparser(parser):
    parser.description = 'help for DAMON monitoring results filtering options'
    return parser



# SPDX-License-Identifier: GPL-2.0

"""
Start DAMON with given parameters.
"""

import os
import signal
import time

import _damo_records
import _damon
import _damon_args

def module_running(module_name):
    param_dir = os.path.join('/sys/module', module_name, 'parameters')
    for param_name in ['enabled', 'enable']:
        param_file = os.path.join(param_dir, param_name)
        if os.path.isfile(param_file):
            with open(param_file, 'r') as f:
                return f.read().strip() == 'Y'
    return False

def module_disable(module_name):
    param_dir = os.path.join('/sys/module', module_name, 'parameters')
    for param_name in ['enabled', 'enable']:
        param_file = os.path.join(param_dir, param_name)
        if os.path.isfile(param_file):
            with open(param_file, 'w') as f:
                f.write('N')
                return

def handle_modules():
    for module in os.listdir('/sys/module'):
        if not module.startswith('damon_'):
            continue
        if not module_running(module):
            continue
        print('Cannot turn on damon since %s is running.  '
              'You should disable it first.' % module)
        answer = input('May I disable it for you? [Y/n] ')
        if answer.lower() == 'n':
            print('Ok, see you later')
            exit(1)
        print('Ok, disabling it')
        module_disable(module)
        print('Disabled it.  Continue starting DAMON')

def sighandler(signum, frame):
    print('\nsingal %s received' % signum)
    exit(0)

def main(args):
    _damon.ensure_root_and_initialized(args)
    handle_modules()

    err, kdamonds = _damon_args.turn_damon_on(args)
    if err:
        print('could not turn on damon (%s)' % err)
        exit(1)

    if args.include_child_tasks is True:
        signal.signal(signal.SIGINT, sighandler)
        signal.signal(signal.SIGTERM, sighandler)
        print('Continue monitoring child tasks and updating DAMON targets')
        print('Press Ctrl+C to stop')
        while True:
            _damo_records.add_childs_target(kdamonds)
            time.sleep(3)

def set_argparser(parser):
    _damon_args.set_argparser(parser, add_record_options=False, min_help=True)
    parser.add_argument('--include_child_tasks', action='store_true',
                        help='add child tasks as monitoring target')
    parser.description = 'Start DAMON with specified parameters'
    return parser



# SPDX-License-Identifier: GPL-2.0

import os
import setuptools
os.sys.path.insert(0, os.path.join(os.path.dirname(os.path.abspath(__file__)), "src"))

from damo import damo_version

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="damo",
    version=damo_version.__version__,
    author="SeongJae Park",
    author_email="sj@kernel.org",
    description="DAMON user-space tool",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/damonitor/damo",
    project_urls={
        "Bug Tracker": "https://github.com/damonitor/damo/issues",
        "DAMON": "https://damonitor.github.io",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v2 (GPLv2)",
        "Operating System :: POSIX :: Linux",
    ],
    package_dir={"": "src"},
    packages=setuptools.find_packages(where="src"),
    entry_points = {
        "console_scripts": ["damo=damo.damo:main"],
    },
    python_requires=">=3.6",
)



# SPDX-License-Identifier: GPL-2.0

"""
Contains core functions for DAMON sysfs control.
"""

import os
import time

import _damo_fmt_str
import _damo_fs
import _damon

sysfs_root = None

def get_sysfs_root():
    global sysfs_root

    if sysfs_root is None:
        sysfs_root = _damo_fs.dev_mount_point('sysfs')
    return sysfs_root

def get_kdamonds_dir():
    '''Returns None if sysfs is not mounted'''
    if get_sysfs_root() is None:
        return None
    return os.path.join(get_sysfs_root(), 'kernel/mm/damon/admin/kdamonds')

def supported():
    kdamonds_dir = get_kdamonds_dir()
    if kdamonds_dir is None:
        return False
    return os.path.isdir(kdamonds_dir)

def get_state_file_of(kdamond_idx):
    return os.path.join(get_kdamonds_dir(), '%s' % kdamond_idx, 'state')

def __write_state_file(kdamond_idxs, command):
    'Return error'
    err = None
    for kdamond_idx in kdamond_idxs:
        err = _damo_fs.write_file(get_state_file_of(kdamond_idx), command)
        if err != None:
            break
    return err

def turn_damon_on(kdamonds_idxs):
    # In case of vaddr, too early monitoring shows unstable mapping changes.
    # Give the process a time to have stable memory mapping.
    time.sleep(0.5)
    return __write_state_file(kdamonds_idxs, 'on')

def turn_damon_off(kdamonds_idxs):
    return __write_state_file(kdamonds_idxs, 'off')

def is_kdamond_running(kdamond_idx):
    content, err = _damo_fs.read_file(get_state_file_of(kdamond_idx))
    if err != None:
        print(err)
        return False
    return content.strip() == 'on'

'Return error'
def update_tuned_intervals(kdamond_idxs):
    return __write_state_file(kdamond_idxs, 'update_tuned_intervals')

'Return error'
def update_schemes_stats(kdamond_idxs):
    return __write_state_file(kdamond_idxs, 'update_schemes_stats')

'Return error'
def update_schemes_tried_bytes(kdamond_idxs):
    err = __write_state_file(kdamond_idxs, 'update_schemes_tried_bytes')
    if err != None:
        err = '%s (maybe schemes_tried_regions_sz not supported?)' % err
    return err

'Return error'
def update_schemes_tried_regions(kdamond_idxs):
    err = __write_state_file(kdamond_idxs, 'update_schemes_tried_regions')
    if err != None:
        if not feature_supported('schemes_tried_regions'):
            err = '%s (DAMON feature \'schemes_tried_regions\' is not supported on the current kernel. It is available on kernel version 6.2 and later)' % err
        else:
            err = '%s (concurrent threads doing writes)' % err
    return err

'Return error'
def update_schemes_quota_effective_bytes(kdamond_idxs):
    err = __write_state_file(kdamond_idxs, 'update_schemes_effective_quotas')
    if err != None:
        err = '%s (maybe schemes_effective_quotas not supported?)' % err
    return err

# for stage_kdamonds

def write_filter_dir(dir_path, filter_):
    err = _damo_fs.write_file(
            os.path.join(dir_path, 'type'), filter_.filter_type)
    if err is not None:
        # todo: make error message more detailed.
        # anon/memcg are merged in 6.3-rc1
        # addr/target are merged in 6.6-rc1
        return err

    if filter_.allow is True:
        if os.path.isfile(os.path.join(dir_path, 'allow')):
            err = _damo_fs.write_file(os.path.join(dir_path, 'allow'),
                                      'Y' if filter_.allow else 'N')
        # pass file has renamed to allow during development
        elif os.path.isfile(os.path.join(dir_path, 'pass')):
            err = _damo_fs.write_file(os.path.join(dir_path, 'pass'),
                                      'Y' if filter_.allow else 'N')
        else:
            return 'kernel is not supporting allow_filter'
        if err is not None:
            return err

    if filter_.memcg_path is not None:
        err = _damo_fs.write_file(
                os.path.join(dir_path, 'memcg_path'), filter_.memcg_path)
        if err is not None:
            return err

    if filter_.hugepage_size is not None:
        err = _damo_fs.write_file(
                os.path.join(dir_path, 'min'),
                '%d' % filter_.hugepage_size[0])
        if err is not None:
            return err

        err = _damo_fs.write_file(
                os.path.join(dir_path, 'max'),
                '%d' % filter_.hugepage_size[1])
        if err is not None:
            return err

    if filter_.address_range is not None:
        err = _damo_fs.write_file(
                os.path.join(dir_path, 'addr_start'),
                '%d' % filter_.address_range.start)
        if err is not None:
            return err

        err = _damo_fs.write_file(
                os.path.join(dir_path, 'addr_end'),
                '%d' % filter_.address_range.end)
        if err is not None:
            return err

    if filter_.damon_target_idx is not None:
        err = _damo_fs.write_file(
                os.path.join(dir_path, 'damon_target_idx'),
                '%d' % filter_.damon_target_idx)
        if err is not None:
            return err

    return _damo_fs.write_file(os.path.join(dir_path, 'matching'),
                               'Y' if filter_.matching else 'N')

def ensure_nr_file_for(file_path, list_):
    content, err = _damo_fs.read_file(file_path)
    if err is not None:
        return err
    current_nr = int(content)
    desired_nr = len(list_)
    if current_nr == desired_nr:
        return None
    return _damo_fs.write_file(file_path, '%d' % desired_nr)

def write_filters_dir(dir_path, filters):
    # filters merged in v6.3-rc1
    if not os.path.isdir(dir_path):
        if len(filters) == 0:
            return None
        return 'the kernel is not supporting filters'

    err = ensure_nr_file_for(os.path.join(dir_path, 'nr_filters'), filters)
    if err is not None:
        return err

    for idx, filter_ in enumerate(filters):
        err = write_filter_dir(os.path.join(dir_path, '%d' % idx), filter_)
        if err is not None:
            return err
    return None

def is_core_filter(damos_filter):
    return damos_filter.filter_type in ['addr', 'target']

def write_core_ops_filters_dirs(scheme_dir_path, filters):
    core_filters = [f for f in filters if is_core_filter(f)]
    err = write_filters_dir(os.path.join(scheme_dir_path, 'core_filters'),
                            core_filters)
    if err is not None:
        return err

    ops_filters = [f for f in filters if not is_core_filter(f)]
    err = write_filters_dir(os.path.join(scheme_dir_path, 'ops_filters'),
                            ops_filters)
    if err is not None:
        return err

def write_filters_dirs(scheme_dir_path, filters):
    if os.path.isdir(os.path.join(scheme_dir_path, 'core_filters')):
        return write_core_ops_filters_dirs(scheme_dir_path, filters)
    return write_filters_dir(os.path.join(scheme_dir_path, 'filters'), filters)

def write_watermarks_dir(dir_path, wmarks):
    if wmarks is None:
        # TODO: ensure wmarks is not None
        return None
    err = _damo_fs.write_file(os.path.join(dir_path, 'metric'), wmarks.metric)
    if err is not None:
        return err

    err = _damo_fs.write_file(
            os.path.join(dir_path, 'interval_us'), '%d' % wmarks.interval_us)
    if err is not None:
        return err

    err = _damo_fs.write_file(
            os.path.join(dir_path, 'high'), '%d' % wmarks.high_permil)
    if err is not None:
        return err
    err = _damo_fs.write_file(
            os.path.join(dir_path, 'mid'), '%d' % wmarks.mid_permil)
    if err is not None:
        return err
    return _damo_fs.write_file(
            os.path.join(dir_path, 'low'), '%d' % wmarks.low_permil)

def write_quota_goal_dir(dir_path, goal):
    # goal metric is wip as of 6.8-rc4 days.
    if (not os.path.isfile(os.path.join(dir_path, 'target_metric')) and
        goal.metric != 'user_input'):
        return 'the kernel is not supporting quota goal metric'

    if os.path.isfile(os.path.join(dir_path, 'target_metric')):
        err = _damo_fs.write_file(os.path.join(dir_path, 'target_metric'),
                                  '%s' % goal.metric)
        if err is not None:
            return err

    if goal.has_nid():
        err = _damo_fs.write_file(os.path.join(dir_path, 'nid'), '%d' % goal.nid)
        if err is not None:
            return err

    err = _damo_fs.write_file(
            os.path.join(dir_path, 'target_value'),
            '%d' % goal.target_value)
    if err is not None:
        return err

    return _damo_fs.write_file(
            os.path.join(dir_path, 'current_value'),
            '%d' % goal.current_value)

def write_quota_goals_dir(dir_path, goals):
    # goals dir has merged in 6.8-rc1
    if not os.path.isdir(dir_path):
        if len(goals) == 0:
            return None
        return 'the kernel is not supporting schemes quota goals'

    err = ensure_nr_file_for(os.path.join(dir_path, 'nr_goals'), goals)
    if err is not None:
        return err

    for idx, goal in enumerate(goals):
        err = write_quota_goal_dir(os.path.join(dir_path, '%d' % idx), goal)
        if err is not None:
            return err
    return None

def write_quota_weights_dir(dir_path, quotas):
    err = _damo_fs.write_file(os.path.join(dir_path, 'sz_permil'),
                              '%d' % quotas.weight_sz_permil)
    if err is not None:
        return err

    err = _damo_fs.write_file(os.path.join(dir_path, 'nr_accesses_permil'),
                              '%d' % quotas.weight_nr_accesses_permil)
    if err is not None:
        return err

    return _damo_fs.write_file(os.path.join(dir_path, 'age_permil'),
                              '%d' % quotas.weight_age_permil)

def write_quotas_dir(dir_path, quotas):
    err = _damo_fs.write_file(
            os.path.join(dir_path, 'ms'), '%d' % quotas.time_ms)
    if err is not None:
        return err

    err = _damo_fs.write_file(
            os.path.join(dir_path, 'bytes'), '%d' % quotas.sz_bytes)
    if err is not None:
        return err

    err = _damo_fs.write_file(os.path.join(dir_path, 'reset_interval_ms'),
                              '%d' % quotas.reset_interval_ms)
    if err is not None:
        return err

    err = write_quota_weights_dir(os.path.join(dir_path, 'weights'), quotas)
    if err is not None:
        return err

    return write_quota_goals_dir(os.path.join(dir_path, 'goals'), quotas.goals)

def write_ulong(file_path, val):
    return _damo_fs.write_file(
            file_path, '%d' % min(val, _damo_fmt_str.ulong_max))

def write_scheme_access_pattern_dir(dir_path, pattern):
    err = write_ulong(
            os.path.join(dir_path, 'sz', 'min'), pattern.sz_bytes[0])
    if err is not None:
        return err

    err = write_ulong(
            os.path.join(dir_path, 'sz', 'max'), pattern.sz_bytes[1])
    if err is not None:
        return err

    err = write_ulong(os.path.join(dir_path, 'nr_accesses', 'min'),
                      pattern.nr_acc_min_max[0].samples)
    if err is not None:
        return err

    err = write_ulong(os.path.join(dir_path, 'nr_accesses', 'max'),
                      pattern.nr_acc_min_max[1].samples)
    if err is not None:
        return err

    err = write_ulong(os.path.join(dir_path, 'age', 'min'),
                      pattern.age_min_max[0].aggr_intervals)
    if err is not None:
        return err
    return write_ulong(os.path.join(dir_path, 'age', 'max'),
                       pattern.age_min_max[1].aggr_intervals)

def write_scheme_dir(dir_path, scheme):
    err = write_scheme_access_pattern_dir(
            os.path.join(dir_path, 'access_pattern'), scheme.access_pattern)
    if err is not None:
        return err

    err = _damo_fs.write_file(os.path.join(dir_path, 'action'), scheme.action)
    if err is not None:
        return err

    err = write_quotas_dir(os.path.join(dir_path, 'quotas'), scheme.quotas)
    if err is not None:
        return err

    err = write_watermarks_dir(
            os.path.join(dir_path, 'watermarks'), scheme.watermarks)
    if err is not None:
        return err

    if scheme.target_nid is not None:
        err = _damo_fs.write_file(os.path.join(dir_path,'target_nid'), '%s' % scheme.target_nid)
        if err is not None:
            return err

    err = write_filters_dirs(dir_path, scheme.filters)
    if err is not None:
        return err

    apply_interval_file = os.path.join(dir_path, 'apply_interval_us')
    # schemes apply interval is merged in v6.7-rc1
    if os.path.isfile(apply_interval_file):
        err = _damo_fs.write_file(apply_interval_file,
                                  '%d' % scheme.apply_interval_us)
        if err is not None:
            return err
    else:
        if scheme.apply_interval_us:
            return 'the kernel is not supporting schemes apply interval'
    return None

def write_schemes_dir(dir_path, schemes):
    err = ensure_nr_file_for(os.path.join(dir_path, 'nr_schemes'), schemes)
    if err is not None:
        return err

    for idx, scheme in enumerate(schemes):
        err = write_scheme_dir(os.path.join(dir_path, '%d' % idx), scheme)
        if err is not None:
            return err

def write_target_region_dir(dir_path, region):
    err = _damo_fs.write_file(
            os.path.join(dir_path, 'start'), '%d' % region.start)
    if err is not None:
        return err

    return _damo_fs.write_file(
            os.path.join(dir_path, 'end'), '%d' % region.end)

def write_target_regions_dir(dir_path, regions):
    err = ensure_nr_file_for(os.path.join(dir_path, 'nr_regions'), regions)
    if err is not None:
        return err

    for idx, region in enumerate(regions):
        err = write_target_region_dir(
                os.path.join(dir_path, '%d' % idx), region)
        if err is not None:
            return err
    return None

def write_target_dir(dir_path, target):
    if target.pid is not None:
        err = _damo_fs.write_file(
                os.path.join(dir_path, 'pid_target'), '%s' % target.pid)
        if err is not None:
            return err

    return write_target_regions_dir(
            os.path.join(dir_path, 'regions'), target.regions)


def write_targets_dir(dir_path, targets):
    err = ensure_nr_file_for(os.path.join(dir_path, 'nr_targets'), targets)
    if err is not None:
        return err

    for idx, target in enumerate(targets):
        err = write_target_dir(os.path.join(dir_path, '%d' % idx), target)
        if err is not None:
            return err
    return None

def write_monitoring_intervals_goal_dir(dir_path, goal):
    err = _damo_fs.write_file(os.path.join(dir_path, 'access_bp'), '%d' %
                              goal.access_bp)
    if err is not None:
        return err
    err = _damo_fs.write_file(os.path.join(dir_path, 'aggrs'), '%d' %
                              goal.aggrs)
    if err is not None:
        return err
    err = _damo_fs.write_file(os.path.join(dir_path, 'min_sample_us'), '%d' %
                              goal.min_sample_us)
    if err is not None:
        return err
    err = _damo_fs.write_file(os.path.join(dir_path, 'max_sample_us'), '%d' %
                              goal.max_sample_us)
    if err is not None:
        return err

def write_monitoring_attrs_dir(dir_path, context):
    err = _damo_fs.write_file(
            os.path.join(dir_path, 'intervals', 'sample_us'),
            '%d' % context.intervals.sample)
    if err is not None:
        return err

    err = _damo_fs.write_file(
            os.path.join(dir_path, 'intervals', 'aggr_us'),
            '%d' % context.intervals.aggr)
    if err is not None:
        return err

    err = _damo_fs.write_file(
            os.path.join(dir_path, 'intervals', 'update_us'),
            '%d' % context.intervals.ops_update)
    if err is not None:
        return err

    intervals_goal_path = os.path.join(dir_path, 'intervals', 'intervals_goal')
    if os.path.isdir(intervals_goal_path):
        err = write_monitoring_intervals_goal_dir(
                intervals_goal_path, context.intervals.intervals_goal)
        if err is not None:
            return err

    err = _damo_fs.write_file(
            os.path.join(dir_path, 'nr_regions', 'min'),
            '%d' % context.nr_regions.minimum)
    if err is not None:
        return err

    return _damo_fs.write_file(
            os.path.join(dir_path, 'nr_regions', 'max'),
            '%d' % context.nr_regions.maximum)

def write_context_dir(dir_path, context):
    err = _damo_fs.write_file(os.path.join(dir_path, 'operations'),
                              context.ops)
    if err is not None:
        return err

    if context.addr_unit is not None:
        err = _damo_fs.write_file(os.path.join(dir_path, 'addr_unit'),
                                  context.addr_unit)
        if err is not None:
            return err

    err = write_monitoring_attrs_dir(
            os.path.join(dir_path, 'monitoring_attrs'), context)
    if err is not None:
        return err

    err = write_targets_dir(
            os.path.join(dir_path, 'targets'), context.targets)
    if err is not None:
        return err

    for scheme in context.schemes:
        scheme.access_pattern = scheme.access_pattern.converted_for_units(
                _damon.unit_samples, _damon.unit_aggr_intervals,
                context.intervals)
    return write_schemes_dir(
            os.path.join(dir_path, 'schemes'), context.schemes)

def write_contexts_dir(dir_path, contexts):
    err = ensure_nr_file_for(os.path.join(dir_path, 'nr_contexts'), contexts)
    if err is not None:
        return err

    for idx, context in enumerate(contexts):
        err = write_context_dir(
                os.path.join(dir_path, '%d' % idx), context)
        if err is not None:
            return err

def write_kdamonds_dir(dir_path, kdamonds):
    err = ensure_nr_file_for(os.path.join(dir_path, 'nr_kdamonds'), kdamonds)
    if err:
        return err

    for idx, kdamond in enumerate(kdamonds):
        err = write_contexts_dir(
                os.path.join(dir_path, '%d' % idx, 'contexts'),
                kdamond.contexts)
        if err is not None:
            return err

def stage_kdamonds(kdamonds):
    """Write DAMON parameters for kdamonds to the sysfs files.

    Args:
        kdamonds: A list of _damon.Kdamond objects.

    Returns:
        None for success, an error string if failed.
    """
    # Assume caller checked supported()
    return write_kdamonds_dir(get_kdamonds_dir(), kdamonds)

# for current_kdamonds()

def numbered_dirs_content(files_content, nr_filename):
    nr_dirs = int(files_content[nr_filename])
    number_dirs = []
    for i in range(nr_dirs):
        number_dirs.append(files_content['%d' % i])
    return number_dirs

def number_sorted_dirs(files_content):
    number_dirs = {}
    for filename, content in files_content.items():
        try:
            nr = int(filename)
        except:
            continue
        if type(content) != dict:
            continue
        number_dirs[nr] = content
    sorted_numbers = sorted(number_dirs.keys())
    return [number_dirs[nr] for nr in sorted_numbers]

def files_content_to_access_pattern(files_content):
    return _damon.DamosAccessPattern(
            [int(files_content['sz']['min']),
                int(files_content['sz']['max'])],
            [int(files_content['nr_accesses']['min']),
                int(files_content['nr_accesses']['max'])],
            _damon.unit_samples, # nr_accesses_unit
            [int(files_content['age']['min']),
                int(files_content['age']['max'])],
            _damon.unit_aggr_intervals) # age_unit

def files_content_to_quota_goals(files_content):
    goals = []
    for goal_kv in number_sorted_dirs(files_content):
        if 'target_metric' in goal_kv:
            goals.append(
                    _damon.DamosQuotaGoal(
                        metric=goal_kv['target_metric'].strip(),
                        nid=goal_kv['nid'] if 'nid' in goal_kv else None,
                        target_value=goal_kv['target_value'],
                        current_value=goal_kv['current_value']))
        else:
            goals.append(
                    _damon.DamosQuotaGoal(
                        target_value=goal_kv['target_value'],
                        current_value=goal_kv['current_value']))
    return goals

def files_content_to_quotas(files_content):
    return _damon.DamosQuotas(
            int(files_content['ms']),
            int(files_content['bytes']),
            int(files_content['reset_interval_ms']),
            [int(files_content['weights']['sz_permil']),
                int(files_content['weights']['nr_accesses_permil']),
                int(files_content['weights']['age_permil'])],
            files_content_to_quota_goals(files_content['goals'])
            if 'goals' in files_content else [],
            int(files_content['effective_bytes'])
            if 'effective_bytes' in files_content else 0)

def files_content_to_watermarks(files_content):
    return _damon.DamosWatermarks(
            files_content['metric'].strip(),
            int(files_content['interval_us']),
            int(files_content['high']),
            int(files_content['mid']),
            int(files_content['low']))

def files_content_to_damos_filter(files_content):
    allow = False
    if 'allow' in files_content:
        allow = files_content['allow'].strip()
    # the file name has changed from pass to allow after v1 patch
    elif 'pass' in files_content:
        allow = files_content['pass'].strip()

    hugepage_size=None
    if 'min' in files_content and 'max' in files_content:
        hugepage_size = [_damo_fmt_str.text_to_bytes(x)
                         for x in [files_content['min'], files_content['max']]]

    return _damon.DamosFilter(
        files_content['type'].strip(),
        files_content['matching'].strip(),
        allow,
        files_content['memcg_path'].strip(),
        _damon.DamonRegion(files_content['addr_start'].strip(),
                           files_content['addr_end'].strip())
        if 'addr_start' in files_content and 'addr_end' in files_content
        else None,
        files_content['damon_target_idx']
        if 'damon_target_idx' in files_content else None,
        hugepage_size=hugepage_size)

def files_content_to_damos_filters(scheme_files_content):
    filters = []
    if 'core_filters' in scheme_files_content:
        filters += [files_content_to_damos_filter(filter_kv)
                    for filter_kv in numbered_dirs_content(
                        scheme_files_content['core_filters'], 'nr_filters')]
        filters += [files_content_to_damos_filter(filter_kv)
                    for filter_kv in numbered_dirs_content(
                        scheme_files_content['ops_filters'], 'nr_filters')]
    filters += [files_content_to_damos_filter(filter_kv)
                for filter_kv in numbered_dirs_content(
                    scheme_files_content['filters'], 'nr_filters')]
    return filters

def files_content_to_damos_stats(files_content):
    return _damon.DamosStats(
            int(files_content['nr_tried']),
            int(files_content['sz_tried']),
            int(files_content['nr_applied']),
            int(files_content['sz_applied']),
            int(files_content['sz_ops_filter_passed']
                if 'sz_ops_filter_passed' in files_content else 0),
            int(files_content['qt_exceeds']))

def files_content_to_damos_tried_regions(files_content):
    return [_damon.DamonRegion(
            int(kv['start']), int(kv['end']),
            int(kv['nr_accesses']), _damon.unit_samples,
            int(kv['age']), _damon.unit_aggr_intervals,
            int(kv['sz_filter_passed'])
            if 'sz_filter_passed' in kv else None,
            ) for kv in number_sorted_dirs(files_content)]

def files_content_to_scheme(files_content):
    return _damon.Damos(
            files_content_to_access_pattern(files_content['access_pattern']),
            files_content['action'].strip(),
            files_content['target_nid'].strip()
                if 'target_nid' in files_content else None,
            files_content['apply_interval_us'].strip()
                if 'apply_interval_us' in files_content else None,
            files_content_to_quotas(files_content['quotas']),
            files_content_to_watermarks(files_content['watermarks']),
            files_content_to_damos_filters(files_content)
                if 'filters' in files_content else [],
            files_content_to_damos_stats(files_content['stats']),
            files_content_to_damos_tried_regions(
                files_content['tried_regions'])
                if 'tried_regions' in files_content else [],
            files_content['tried_regions']['total_bytes']
                if 'tried_regions' in files_content and
                    'total_bytes' in files_content['tried_regions'] else None)

def files_content_to_regions(files_content):
    return [_damon.DamonRegion(
            int(kv['start']), int(kv['end']))
            for kv in numbered_dirs_content(files_content, 'nr_regions')]

def files_content_to_target(files_content):
    try:
        pid = int(files_content['pid_target'])
    except:
        pid = None
    regions = files_content_to_regions(files_content['regions'])
    return _damon.DamonTarget(pid, regions)

def files_content_to_context(files_content):
    mon_attrs_content = files_content['monitoring_attrs']
    intervals_content = mon_attrs_content['intervals']
    if 'intervals_goal' in intervals_content:
        kvpairs = intervals_content['intervals_goal']
        intervals_goal = _damon.DamonIntervalsGoal(
                int(kvpairs['access_bp']), int(kvpairs['aggrs']),
                int(kvpairs['min_sample_us']), int(kvpairs['max_sample_us']))
    else:
        intervals_goal = _damon.DamonIntervalsGoal()
    intervals = _damon.DamonIntervals(
            int(intervals_content['sample_us']),
            int(intervals_content['aggr_us']),
            int(intervals_content['update_us']),
            intervals_goal)
    nr_regions_content = mon_attrs_content['nr_regions']
    nr_regions = _damon.DamonNrRegionsRange(
            int(nr_regions_content['min']),
            int(nr_regions_content['max']))
    ops = files_content['operations'].strip()
    if 'addr_unit' in files_content:
        addr_unit = files_content['addr_unit']
    else:
        addr_unit = None

    targets_content = files_content['targets']
    targets = [files_content_to_target(content)
            for content in numbered_dirs_content(
                targets_content, 'nr_targets')]

    schemes_content = files_content['schemes']
    schemes = [files_content_to_scheme(content)
            for content in numbered_dirs_content(
                schemes_content, 'nr_schemes')]

    return _damon.DamonCtx(ops, targets, intervals, nr_regions, schemes,
                           addr_unit=addr_unit)

def files_content_to_kdamond(files_content):
    contexts_content = files_content['contexts']
    contexts = [files_content_to_context(content)
            for content in numbered_dirs_content(
                contexts_content, 'nr_contexts')]
    state = files_content['state'].strip()
    pid = files_content['pid'].strip()
    return _damon.Kdamond(state, pid, contexts)

def files_content_to_kdamonds(files_contents):
    return [files_content_to_kdamond(content)
            for content in numbered_dirs_content(
                files_contents, 'nr_kdamonds')]

def current_kdamonds():
    # Assume caller checked supported()
    return files_content_to_kdamonds(_damo_fs.read_files(get_kdamonds_dir()))

def get_nr_kdamonds_file():
    return os.path.join(get_kdamonds_dir(), 'nr_kdamonds')

def nr_kdamonds():
    nr_kdamonds, err = _damo_fs.read_file(get_nr_kdamonds_file())
    if err != None:
        raise Exception('nr_kdamonds_file read fail (%s)' % err)
    return int(nr_kdamonds)

def commit_staged(kdamond_idxs):
    for kdamond_idx in kdamond_idxs:
        err = _damo_fs.write_file(get_state_file_of(kdamond_idx), 'commit')
        if err != None:
            return err
    return None

def commit_quota_goals(kdamond_idxs):
    for kdamond_idx in kdamond_idxs:
        err = _damo_fs.write_file(get_state_file_of(kdamond_idx),
                'commit_schemes_quota_goals')
        if err != None:
            return err

# features

feature_supports = None

def feature_supported(feature):
    if feature_supports == None:
        update_supported_features()
    return feature_supports[feature]

# sysfs was merged in v5.18-rc1
features_sysfs_support_from_begining = [
        'schemes',
        'init_regions',
        'vaddr',
        'paddr',
        'init_regions_target_idx',
        'schemes_speed_limit',
        'schemes_quotas',
        'schemes_prioritization',
        'schemes_wmarks',
        'schemes_stat_succ',
        'schemes_stat_qt_exceed',
        ]

def kdamond_dir_of(kdamond_idx):
    return os.path.join(get_kdamonds_dir(), '%s' % kdamond_idx)

def ctx_dir_of(kdamond_idx, context_idx):
    return os.path.join(
            kdamond_dir_of(kdamond_idx), 'contexts', '%s' % context_idx)

def schemes_dir_of(kdamond_idx, context_idx):
    return os.path.join(ctx_dir_of(kdamond_idx, context_idx), 'schemes')

def _avail_ops():
    '''Assumes called by update_supported_features() assuming one scheme.
    Returns available ops input and error'''
    avail_ops = []
    avail_operations_filepath = os.path.join(ctx_dir_of(0, 0),
            'avail_operations')
    if not os.path.isfile(avail_operations_filepath):
        operations_filepath = os.path.join(ctx_dir_of(0, 0), 'operations')
        for ops in ['vaddr', 'paddr', 'fvaddr']:
            err = _damo_fs.write_file(operations_filepath, ops)
            if err != None:
                avail_ops.append(ops)
        return avail_ops, None

    content, err = _damo_fs.read_file(avail_operations_filepath)
    if err != None:
        return None, err
    return content.strip().split(), None

def scheme_dir_of(kdamond_idx, context_idx, scheme_idx):
    return os.path.join(
            schemes_dir_of(kdamond_idx, context_idx), '%s' % scheme_idx)

def scheme_tried_regions_dir_of(kdamond_idx, context_idx, scheme_idx):
    return os.path.join(
            scheme_dir_of(kdamond_idx, context_idx, scheme_idx),
            'tried_regions')

def infer_damon_version():
    orig_kdamonds = current_kdamonds()

    kdamonds = [
            _damon.Kdamond(
                state=None, pid=None, contexts=[
                    _damon.DamonCtx(
                        schemes=[
                            _damon.Damos(
                                filters=[_damon.DamosFilter('young', True)]
                                )])])]
    err = stage_kdamonds(kdamonds)
    if err is None:
        if os.path.isfile(os.path.join(ctx_dir_of(0, 0), 'addr_unit')):
            version = '>v6.15'
        elif os.path.isdir(
                os.path.join(ctx_dir_of(0, 0),
                             'monitoring_attrs', 'intervals', 'intervals_goal')):
            version = '>v6.14'
        elif os.path.isfile(os.path.join(scheme_dir_of(0, 0, 0), 'stats',
                                       'sz_ops_filter_passed')):
            version = '>v6.13'
        elif os.path.isfile(os.path.join(scheme_dir_of(0, 0, 0), 'target_nid')):
            version = '>=v6.11'
        else:
            version = 'v6.10'
        err = stage_kdamonds(orig_kdamonds)
        return version

    kdamonds[0].contexts[0].schemes[0].filters = []
    err = stage_kdamonds(kdamonds)

    if os.path.isfile(os.path.join(scheme_dir_of(0, 0, 0), 'quotas',
                                   'effective_bytes')):
        stage_kdamonds(orig_kdamonds)
        return '6.9'

    if os.path.isdir(os.path.join(scheme_dir_of(0, 0, 0), 'quotas', 'goals')):
        stage_kdamonds(orig_kdamonds)
        return '6.8'

    if os.path.isfile(os.path.join(scheme_dir_of(0, 0, 0), 'apply_interval_us')):
        stage_kdamonds(orig_kdamonds)
        return '6.7'

    if os.path.isfile(os.path.join(scheme_tried_regions_dir_of(0, 0, 0),
            'total_bytes')):
        stage_kdamonds(orig_kdamonds)
        return '6.6'

    if os.path.isdir(os.path.join(scheme_dir_of(0, 0, 0), 'filters')):
        stage_kdamonds(orig_kdamonds)
        return '6.3'

    if os.path.isdir(scheme_tried_regions_dir_of(0, 0, 0)):
        stage_kdamonds(orig_kdamonds)
        return '6.2'

    stage_kdamonds(orig_kdamonds)
    return '<6.2'

def update_supported_features():
    global feature_supports

    if feature_supports != None:
        return None
    feature_supports = {x: False for x in _damon.features}

    if not supported():
        return 'damon sysfs not supported'
    for feature in features_sysfs_support_from_begining:
        feature_supports[feature] = True

    orig_kdamonds = current_kdamonds()
    kdamonds_for_feature_check = [
            _damon.Kdamond(
                state=None, pid=None, contexts=[
                    _damon.DamonCtx(
                        schemes=[_damon.Damos()])])]
    err = stage_kdamonds(kdamonds_for_feature_check)
    if err is not None:
        print('staging feature check purpose kdamond failed')
        stage_kdamonds(orig_kdamonds)
        exit(1)

    if os.path.isdir(scheme_tried_regions_dir_of(0, 0, 0)):
        feature_supports['schemes_tried_regions'] = True

    if os.path.isfile(os.path.join(scheme_tried_regions_dir_of(0, 0, 0),
            'total_bytes')):
        feature_supports['schemes_tried_regions_sz'] = True
        # address and target filter types are added in v6.6-rc1, together with
        # schemes_tried_regions_sz
        feature_supports['schemes_filters_addr'] = True
        feature_supports['schemes_filters_target'] = True

    if os.path.isdir(os.path.join(scheme_dir_of(0, 0, 0), 'filters')):
        feature_supports['schemes_filters'] = True
        # anon and memcg were supported from the beginning
        feature_supports['schemes_filters_anon'] = True
        feature_supports['schemes_filters_memcg'] = True
        kdamonds_for_feature_check[0].contexts[0].schemes[0].filters = [
                _damon.DamosFilter('young', True)]
        err = stage_kdamonds(kdamonds_for_feature_check)
        if err is None:
            feature_supports['schemes_filters_young'] = True

    if os.path.isfile(os.path.join(scheme_dir_of(0, 0, 0), 'apply_interval_us')):
        feature_supports['schemes_apply_interval'] = True

    if os.path.isdir(os.path.join(scheme_dir_of(0, 0, 0), 'quotas', 'goals')):
        feature_supports['schemes_quota_goals'] = True

    if os.path.isfile(os.path.join(scheme_dir_of(0, 0, 0), 'quotas',
                                   'effective_bytes')):
        feature_supports['schemes_quota_effective_bytes'] = True
        # goal_metric and goal_some_psi will be merged together with effective bytes.
        feature_supports['schemes_quota_goal_metric'] = True
        feature_supports['schemes_quota_goal_some_psi'] = True

    if os.path.isfile(os.path.join(scheme_dir_of(0, 0, 0), 'target_nid')):
        feature_supports['schemes_migrate'] = True

    if os.path.isfile(
            os.path.join(scheme_dir_of(0, 0, 0),
                         'stats', 'sz_ops_filter_passed')):
        feature_supports['sz_ops_filter_passed'] = True

    if os.path.isfile(
            os.path.join(scheme_dir_of(0, 0, 0), 'filters', '0', 'allow')):
        feature_supports['allow_filter'] = True

    if os.path.isfile(
            os.path.join(scheme_dir_of(0, 0, 0), 'filters', '0', 'min')):
        feature_supports['schemes_filters_hugepage_size'] = True

    if os.path.isdir(
            os.path.join(ctx_dir_of(0, 0),
                         'monitoring_attrs', 'intervals', 'intervals_goal')):
        feature_supports['intervals_goal'] = True

    if os.path.isdir(
            os.path.join(scheme_dir_of(0, 0, 0), 'core_filters')):
        feature_supports['schemes_filters_core_ops_dirs'] = True

        # unmapped and active pages DAMOS filters are merged into v6.15
        # together with core_ops_dirs
        feature_supports['schemes_filters_unmapped'] = True
        feature_supports['schemes_filters_active'] = True

    if feature_supports['schemes_quota_goals'] is True:
        kdamonds_for_feature_check = [
                _damon.Kdamond(
                    state=None, pid=None, contexts=[
                        _damon.DamonCtx(
                            schemes=[_damon.Damos(
                                quotas=_damon.DamosQuotas(
                                    goals=[_damon.DamosQuotaGoal()])
                                )])])]
        err = stage_kdamonds(kdamonds_for_feature_check)
        if err is not None:
            print('staging damos goal feature check purpose kdamond failed')
            stage_kdamonds(orig_kdamonds)
            exit(1)

        if os.path.isfile(
                os.path.join(scheme_dir_of(0, 0, 0), 'quotas', 'goals', '0',
                             'nid')):
            feature_supports['schemes_quota_goal_node_mem_used_free'] = True

    if os.path.isfile(os.path.join(ctx_dir_of(0, 0), 'addr_unit')):
        feature_supports['addr_unit'] = True

    avail_ops, err = _avail_ops()
    if err == None:
        for ops in ['vaddr', 'paddr', 'fvaddr']:
            feature_supports[ops] = ops in avail_ops
    err = stage_kdamonds(orig_kdamonds)
    return err



# SPDX-License-Identifier: GPL-2.0

"""
Update DAMON input parameters.
"""

import _damon
import _damon_args

def main(args):
    _damon.ensure_root_and_initialized(args)

    if not _damon.any_kdamond_running():
        print('DAMON is not turned on')
        exit(1)

    kdamonds, err = _damon_args.commit_kdamonds(args, args.quota_goals_only)
    if err:
        print('tuning failed (%s)' % err)
        exit(1)

def set_argparser(parser):
    parser.description = 'Update DAMON parameters'
    parser.add_argument('--quota_goals_only', action='store_true',
            help='commit quota goals change only')
    _damon_args.set_argparser(parser, add_record_options=False, min_help=True)



# SPDX-License-Identifier: GPL-2.0

"Print out the distribution of the working set sizes of the given trace"

import sys
import tempfile

import _damo_dist
import _damo_fmt_str
import _damo_records

def get_wss_dists(records, acc_thres, sz_thres, do_sort, collapse_targets):
    wss_dists = {}
    for record in records:
        wss_dist = []
        for snapshot in record.snapshots:
            wss = 0
            for r in snapshot.regions:
                # Ignore regions not fulfill working set conditions
                if r.nr_accesses.samples < acc_thres:
                    continue
                if r.size() < sz_thres:
                    continue
                wss += r.size()
            wss_dist.append(wss)
        if do_sort:
            wss_dist.sort(reverse=False)
        wss_dists[record.target_id] = wss_dist
    if collapse_targets is True:
        collapsed_dist = []
        for t, dist in wss_dists.items():
            for idx, wss in enumerate(dist):
                if len(collapsed_dist) <= idx:
                    collapsed_dist.append(wss)
                else:
                    collapsed_dist[idx] += wss
        wss_dists = {0: collapsed_dist}
    return wss_dists

def set_argparser(parser):
    parser.add_argument('--input', '-i', type=str, metavar='<file>',
            default='damon.data', help='input file name')
    parser.add_argument('--range', '-r', type=int, nargs=3,
            metavar=('<start>', '<stop>', '<step>'), default=[0,101,25],
            help='range of wss percentiles to print')
    parser.add_argument('--exclude_samples', type=int, default=20,
            metavar='<# samples>',
            help='number of first samples to be excluded')
    parser.add_argument('--acc_thres', '-t', type=int, default=1,
            metavar='<# accesses>',
            help='minimal number of accesses for treated as working set')
    parser.add_argument('--sz_thres', type=int, default=1,
            metavar='<size>',
            help='minimal size of region for treated as working set')
    parser.add_argument('--work_time', type=int, default=1,
            metavar='<micro-seconds>',
            help='supposed time for each unit of the work')
    parser.add_argument('--sortby', '-s', choices=['time', 'size'],
            help='the metric to be used for the sort of the working set sizes')
    parser.add_argument('--plot', '-p', type=str, metavar='<file>',
            help='plot the distribution to an image file')
    parser.add_argument('--nr_cols_bar', type=int, metavar='<num>',
            default=59,
            help='number of columns that is reserved for wss visualization')
    parser.add_argument('--raw_number', action='store_true',
            help='use machine-friendly raw numbers')
    parser.add_argument('--all_wss', action='store_true',
            help='Do not print percentile but all calculated wss')
    parser.add_argument('--per_target', action='store_false',
                        dest='collapse_targets',
                        help='Report workingset size per monitoring target')
    parser.add_argument('--collapse_targets', action='store_true',
                        help='Collapse targets in the record into one')
    parser.description = 'Show distribution of working set size'

def main(args):
    file_path = args.input
    percentiles = range(args.range[0], args.range[1], args.range[2])
    wss_sort = True
    if args.sortby == 'time':
        wss_sort = False
    raw_number = args.raw_number

    records, err = _damo_records.get_records(record_file=file_path)
    if err != None:
        print('monitoring result file (%s) parsing failed (%s)' %
                (file_path, err))
        exit(1)

    _damo_records.adjust_records(records, args.work_time, args.exclude_samples)
    wss_dists = get_wss_dists(records, args.acc_thres, args.sz_thres, wss_sort,
                              args.collapse_targets)

    if not args.plot:
        for tid, dists in wss_dists.items():
            print('# target_id\t%s' % tid)
            _damo_dist.pr_dists(
                    'wss', dists, percentiles, args.all_wss,
                    _damo_fmt_str.format_sz, raw_number, args.nr_cols_bar)
        return

    tmp_path = tempfile.mkstemp()[1]
    with open(tmp_path, 'w') as f:
        for tid, dists in wss_dists.items():
            f.write('# target_id\t%s\n' % tid)
            f.write( _damo_dist.fmt_dists(
                'wss', dists, percentiles, args.all_wss,
                _damo_fmt_str.format_sz, True, 0))

    xlabel = 'runtime (percent)'
    if wss_sort:
        xlabel = 'percentile'
    err = _damo_dist.plot_dist(tmp_path, args.plot, xlabel,
            'working set size (bytes)')
    if err:
        print('plot failed (%s)' % err)
